# -*- coding: utf-8 -*-
"""
pyee
====

pyee supplies an ``EventEmitter`` object similar to the ``EventEmitter``
from Node.js.


Example
-------

::

    In [1]: from pyee import EventEmitter

    In [2]: ee = EventEmitter()

    In [3]: @ee.on('event')
       ...: def event_handler():
       ...:     print 'BANG BANG'
       ...:

    In [4]: ee.emit('event')
    BANG BANG

    In [5]:


Easy-peasy.
"""

import collections
import contextlib
import functools
import sys


__all__ = [
    'EventEmitter',
]


class EventEmitter(object):
    """
    The EventEmitter class.

    (Special) Events
    ----------------

    'new_listener': Fires whenever a new listener is created. Listeners for
        this event do not fire upon their own creation.

    'error': When emitted raises an Exception by default, behavior can be
        overriden by attaching callback to the event.

    For example::

        @ee.on('error')
        def on_rror(*exc_info):
            logging.error('something bad happened', exc_info=exc_info)

        ee.emit('error', Exception('something blew up'))
    """

    __slots__ = (
        '_events',
    )

    def __init__(self):
        """
        Initializes the emitter.
        """
        self._events = collections.defaultdict(list)

    def on(self, event, f=None):
        """Registers the function ``f`` to the event name ``event``.

        If ``f`` isn't provided, this method returns a function that
        takes ``f`` as a callback; in other words, you can use this method
        as a decorator, like so:

            @ee.on('data')
            def data_handler(data):
                print data

        """
        def _on(f):
            # Add the necessary function
            self.listeners(event).append(f)

            # Return original function so removal works
            return f

        if f is None:
            return _on
        else:
            return _on(f)

    def emit(self, event, *args, **kwargs):
        """
        Emit ``event``, passing ``*args`` to each attached function. Returns
        ``True`` if any functions are attached to ``event``; otherwise returns
        ``False``.

        Example:
            ee.emit('data', '00101001')

        Assuming ``data`` is an attached function, this will call
        ``data('00101001')'``.

        ``error`` is a special type of event. Either a single exception
        instance must be passed or the values from ``sys.exc_info()``. An
        example::

            ee.emit('error', *sys.exc_info())

            my_error = RuntimeError('foobar')

            ee.emit('error', my_error)

        All listeners to an error event will receive a sys.exc_info style
        tuple, even if an exception instance is supplied.

        If the ``error`` event is not handled, the exception is raised inline.
        """
        listeners = self.listeners(event)

        if event == 'error':
            # convert args in to a tuple as returned by sys.exc_info
            args = get_exc_info(*args)

            if not listeners:
                raise args[0], args[1], args[2]

        # Pass the args to each function in the events dict
        for func in listeners:
            func(*args, **kwargs)

        # whether the event was handled.
        return bool(listeners)

    def once(self, event, f=None):
        """The same as ``ee.on``, except that the listener is automatically
        removed after being called.
        """
        def _once(f):
            @functools.wraps(f)
            def g(*args, **kwargs):
                try:
                    self.remove_listener(event, g)
                finally:
                    f(*args, **kwargs)

            return g

        if f is None:
            return lambda f: self.on(event, _once(f))
        else:
            return self.on(event, _once(f))

    def remove_listener(self, event, f):
        """
        Removes the function ``f`` from ``event``.

        Requires that ``f`` is not closed over by ``ee.on``. (In other words,
        it is, unfortunately, not possible to use this with the decorator
        style is.)

        """
        self._events[event].remove(f)

    def remove_all_listeners(self, event=None):
        """
        Remove all listeners attached to ``event``.
        """
        if event is not None:
            self._events[event] = []
        else:
            self._events = None
            self._events = collections.defaultdict(list)

    def listeners(self, event):
        """
        Returns the list of all listeners registered to the ``event``.
        """
        return self._events[event]


@contextlib.contextmanager
def emit_exceptions(emitter, logger, propagate=True, always_log=False,
                    emit=True, skip_types=None):
    """
    A context manager that wraps a chunk of synchronous code and traps any
    exceptions. If an exception is caught and ``emit`` is ``True`` (the
    default), the ``emitter`` will emit an ``error`` event with the
    ``sys.exc_info()`` of the exception that was raised.

    A basic example looks like:

        import logging

        logging.basicConfig()

        ee = EventEmitter()

        @ee.on('error')
        def whoops(*exc_info):
            print('oh dear', exc_info)

        def blow_up():
            raise RuntimeError('foobar')

        with emit_exceptions(ee, logging):
            blow_up()

    The above example will print "oh dear <type 'exceptions.RuntimeError'> ..."
    and also raise the exception inline. This can be useful to ensure that the
    ``error`` event is sent but still have the normal try/except semantics
    still work.

    If an exception is raised when emitting the error, the exception is logged
    but the *original* exception is raised. Keep an eye on your logs ;)

    :param emitter: The event emitter that the `error` event will be
        dispatched from.
    :param logger: The logger that will handle any exceptions.
    :param propagate: Whether the context manager will re-raise the
        exception at `__exit__`.
    :param always_log: If an exception is raised, log the exception even if
        it is sent to the emitter. If `False`, exceptions will only be
        logged if there are no error handlers.
    :param skip_types: A list of exception types to avoid trapping. Exceptions
        of these types will be re-raised.
    """
    try:
        yield
    except (Exception, BaseException) as exc:
        if skip_types and isinstance(exc, skip_types):
            raise

        exc_info = sys.exc_info()
        handled = False

        # we have an exception, now to handle it. if the act of emitting the
        # error event causes an exception to be raised, log it out
        if emit:
            try:
                handled = emitter.emit('error', *exc_info)
            except (Exception, BaseException) as emit_exc:
                if skip_types and isinstance(emit_exc, skip_types):
                    emit_exc = None
                    handled = True

                if emit_exc is not None and emit_exc is not exc:
                    logger.exception(
                        'Exception raised while emitting error event',
                        exc_info=sys.exc_info()
                    )

        if not handled or always_log:
            logger.error('Exception was caught', exc_info=exc_info)

        if propagate:
            raise exc_info[0], exc_info[1], exc_info[2]


def get_exc_info(*args):
    """
    From a list of arguments, return a value that matches the format of
    ``sys.exc_info``.
    """
    if not args:
        raise TypeError('Expected at least one argument')

    if len(args) == 1:
        exc = args[0]

        try:
            if issubclass(exc, (Exception, BaseException)):
                exc = exc()
        except TypeError:
            pass

        if not isinstance(exc, (Exception, BaseException)):
            raise TypeError(
                'An exception instance must be supplied as the first argument '
                '(received:{!r})'.format(exc)
            )

        return exc.__class__, exc, None

    if len(args) != 3:
        raise TypeError(
            'Expected an exception instance or sys.exc_info() tuple'
        )

    exc_type, exc_value, exc_tb = args

    if not isinstance(exc_value, (Exception, BaseException)):
        raise TypeError(
            'exc_info tuple must be a valid exception instance '
            '(received:{!r}, {!r}, {!r})'.format(exc_type, exc_value, exc_tb)
        )

    return exc_type, exc_value, exc_tb



"""
Tests for ``biloba.events``
"""

import unittest
import mock

from biloba import events


class EventEmitterTestCase(unittest.TestCase):
    """
    Tests for ``events.EventEmitter``
    """

    def test_create(self):
        """
        Ensure that an EventEmitter can be instantiated.
        """
        emitter = events.EventEmitter()

        self.assertIsInstance(emitter, events.EventEmitter)

    def test_get_listeners(self):
        """
        ``EventEmitter.listeners`` must return a list for the event type.
        """
        emitter = events.EventEmitter()

        self.assertEqual(emitter.listeners('foobar'), [])

        @emitter.on('foobar')
        def foo_bar():
            pass

        self.assertEqual(emitter.listeners('foobar'), [foo_bar])

    def test_on_as_decorator(self):
        """
        Ensure that ``EventEmitter.on`` registers the listener when used as a
        decorator.
        """
        emitter = events.EventEmitter()

        self.assertEqual(emitter.listeners('foobar'), [])

        @emitter.on('foobar')
        def foo_bar():
            pass

        self.assertEqual(emitter.listeners('foobar'), [foo_bar])

    def test_on_functional(self):
        """
        Ensure that ``EventEmitter.on`` registers the listener when used in a
        purely functional manner
        """
        emitter = events.EventEmitter()

        self.assertEqual(emitter.listeners('foobar'), [])

        def foo_bar():
            pass

        emitter.on('foobar', foo_bar)

        self.assertEqual(emitter.listeners('foobar'), [foo_bar])

    def test_once_decorator(self):
        """
        Ensure that the listener is removed once the event is emitted (added
        as a decorator).
        """
        emitter = events.EventEmitter()

        self.assertEqual(emitter.listeners('foobar'), [])

        @emitter.once('foobar')
        def foo_bar():
            pass

        self.assertEqual(emitter.listeners('foobar'), [foo_bar])

        emitter.emit('foobar')

        self.assertEqual(emitter.listeners('foobar'), [])

    def test_once_functional(self):
        """
        Ensure that the listener is removed once the event is emitted (added in
        a purely functional manner).
        """
        emitter = events.EventEmitter()

        self.assertEqual(emitter.listeners('foobar'), [])

        def foo_bar():
            pass

        func = emitter.once('foobar', foo_bar)

        self.assertEqual(emitter.listeners('foobar'), [func])

        emitter.emit('foobar')

        self.assertEqual(emitter.listeners('foobar'), [])

    def test_once_error(self):
        """
        If the listener raises an exception, it MUST still be removed from the
        emitter.
        """
        emitter = events.EventEmitter()

        self.assertEqual(emitter.listeners('foobar'), [])

        def foo_bar():
            raise RuntimeError

        func = emitter.once('foobar', foo_bar)

        self.assertEqual(emitter.listeners('foobar'), [func])

        with self.assertRaises(RuntimeError):
            emitter.emit('foobar')

        self.assertEqual(emitter.listeners('foobar'), [])

    def test_remove_all_listeners(self):
        """
        ``EventEmitter.remove_all_listeners`` must do as expected.
        """
        emitter = events.EventEmitter()

        emitter.on('foo', lambda: None)
        emitter.on('bar', lambda: None)

        self.assertNotEqual(emitter.listeners('foo'), [])
        self.assertNotEqual(emitter.listeners('bar'), [])

        emitter.remove_all_listeners()

        self.assertEqual(emitter.listeners('foo'), [])
        self.assertEqual(emitter.listeners('bar'), [])

    def test_remove_all_listeners_event(self):
        """
        ``EventEmitter.remove_all_listeners`` with a supplied event must only
        remove those event listeneres.
        """
        emitter = events.EventEmitter()

        emitter.on('foo', lambda: None)
        emitter.on('bar', lambda: None)

        self.assertNotEqual(emitter.listeners('foo'), [])
        self.assertNotEqual(emitter.listeners('bar'), [])

        emitter.remove_all_listeners('foo')

        self.assertEqual(emitter.listeners('foo'), [])
        self.assertNotEqual(emitter.listeners('bar'), [])

    def test_emit_error_instance(self):
        """
        Emitting an ``error`` event must supply the correct args to the
        listener.
        """
        class TestException(Exception):
            pass

        emitter = events.EventEmitter()

        exc = TestException()

        self.executed = False

        @emitter.on('error')
        def on_error(exc_type, exc_value, exc_tb):
            self.assertIs(exc_type, TestException)
            self.assertIs(exc_value, exc)
            self.assertIsNone(exc_tb)

            self.executed = True

        emitter.emit('error', exc)

        self.assertTrue(self.executed)

    def test_emit_error_class(self):
        """
        Emitting an ``error`` event must supply the correct args to the
        listener.
        """
        class TestException(Exception):
            pass

        emitter = events.EventEmitter()

        self.executed = False

        @emitter.on('error')
        def on_error(exc_type, exc_value, exc_tb):
            self.assertIs(exc_type, TestException)
            self.assertIsInstance(exc_value, TestException)
            self.assertIsNone(exc_tb)

            self.executed = True

        emitter.emit('error', TestException)

        self.assertTrue(self.executed)

    def test_emit_error_tuple(self):
        """
        Emitting an ``error`` event must supply the correct args to the
        listener.
        """
        class TestException(Exception):
            pass

        emitter = events.EventEmitter()
        exc = TestException()

        self.executed = False

        @emitter.on('error')
        def on_error(exc_type, exc_value, exc_tb):
            self.assertIs(exc_type, TestException)
            self.assertIs(exc_value, exc)
            self.assertIsNone(exc_tb)

            self.executed = True

        emitter.emit('error', TestException, exc, None)

        self.assertTrue(self.executed)

    def test_emit_error_no_listeners(self):
        """
        Emitting an ``error`` event with no listeners MUST raise the exception.
        """
        class TestException(Exception):
            pass

        emitter = events.EventEmitter()
        exc = TestException()

        with self.assertRaises(TestException) as ctx:
            emitter.emit('error', exc)

        self.assertIs(ctx.exception, exc)


class GetExcInfoTestCase(unittest.TestCase):
    """
    Tests for ``events.get_exc_info``
    """

    def test_missing_args(self):
        """
        Supplying no args must raise a ``TypeError``.
        """
        with self.assertRaises(TypeError) as ctx:
            events.get_exc_info()

        self.assertEqual(
            unicode(ctx.exception),
            u'Expected at least one argument',
        )

    def test_not_an_exception(self):
        """
        Supplying one arg that is not an exception instance must raise a
        ``TypeError``.
        """
        with self.assertRaises(TypeError) as ctx:
            events.get_exc_info('foobar')

        self.assertEqual(
            unicode(ctx.exception),
            u"An exception instance must be supplied as the first argument "
            "(received:'foobar')",
        )

    def test_exception(self):
        """
        Supplying an exception instance must return a tuple in a
        ``sys.exc_info()`` style format.
        """
        class TestException(Exception):
            pass

        exc = TestException()
        ret = events.get_exc_info(exc)

        self.assertTupleEqual(ret, (TestException, exc, None))

    def test_exception_class(self):
        """
        Supplying an exception CLASS must return a tuple in a
        ``sys.exc_info()`` style format.
        """
        class TestException(Exception):
            pass

        ret = events.get_exc_info(TestException)

        self.assertEquals(len(ret), 3)
        exc_type, exc_value, exc_tb = ret

        self.assertIs(exc_type, TestException)
        self.assertIsInstance(exc_value, TestException)
        self.assertIsNone(exc_tb)

    def test_2_args(self):
        """
        Supplying 2 args must raise a ``TypeError``.
        """
        with self.assertRaises(TypeError):
            events.get_exc_info('foo', 'bar')

    def test_4_args(self):
        """
        Supplying 4 args must raise a ``TypeError``.
        """
        with self.assertRaises(TypeError):
            events.get_exc_info('foo', 'bar', 'baz', 'gak')

    def test_exc_tuple(self):
        """
        Supplying 3 args means that a return value from ``sys.exc_info()`` is
        being supplied.
        """
        with self.assertRaises(TypeError) as ctx:
            events.get_exc_info('foo', 'bar', 'baz')

        self.assertTrue(unicode(ctx.exception).startswith(
            u"exc_info tuple must be a valid exception instance"
        ))

    def test_return_sys_exc_info(self):
        """
        Supplying an exc_info tuple must return a tuple in a ``sys.exc_info()``
        style format.
        """
        exc_info = (RuntimeError, RuntimeError(), None)

        self.assertEqual(
            events.get_exc_info(*exc_info),
            exc_info
        )


class EmitExceptionsTestCase(unittest.TestCase):
    """
    Tests for ``events.emit_exceptions``
    """

    def emit_exceptions(self, emitter=None, logger=None, **kwargs):
        return events.emit_exceptions(
            emitter or mock.Mock(),
            logger or mock.Mock(),
            **kwargs
        )

    def test_no_exception(self):
        """
        If there is no exception raise, don't do anything.
        """
        with self.emit_exceptions():
            pass

    def test_skip_types(self):
        """
        If the exception type matches that supplied by omit then the exception
        must be raised.
        """
        class TestException(Exception):
            pass

        emitter = mock.Mock()

        with self.assertRaises(TestException):
            with self.emit_exceptions(emitter, skip_types=(TestException,)):
                raise TestException

    def test_propagate(self):
        """
        If propagate is ``True`` then the exception must be raised.
        """
        class TestException(Exception):
            pass

        emitter = mock.Mock()
        exc = TestException()

        with self.assertRaises(TestException):
            with self.emit_exceptions(emitter, propagate=True):
                raise exc

        mock_emit = emitter.emit
        mock_emit.assert_called_once()

        error, exc_type, exc_value, exc_tb = mock_emit.call_args[0]

        self.assertEqual(error, 'error')
        self.assertIs(exc_type, TestException)
        self.assertIs(exc_value, exc)
        self.assertIsNotNone(exc_tb)

    def test_no_propagate(self):
        """
        If propagate is ``False``, then the exception must NOT be raised.
        """
        class TestException(Exception):
            pass

        emitter = mock.Mock()
        exc = TestException()

        with self.emit_exceptions(emitter, propagate=False):
            raise exc

        mock_emit = emitter.emit
        mock_emit.assert_called_once()

        error, exc_type, exc_value, exc_tb = mock_emit.call_args[0]

        self.assertEqual(error, 'error')
        self.assertIs(exc_type, TestException)
        self.assertIs(exc_value, exc)
        self.assertIsNotNone(exc_tb)

    def test_no_emit(self):
        """
        If ``emit`` is False then the error event must NOT be emitted.
        """
        class TestException(Exception):
            pass

        emitter = mock.Mock()
        logger = mock.Mock()

        ctx = self.emit_exceptions(
            emitter,
            logger,
            emit=False,
            propagate=False
        )

        with ctx:
            raise TestException

        self.assertFalse(emitter.emit.called)
        self.assertTrue(logger.error.called)

    def test_always_log(self):
        """
        If ``always_log`` is True then any exception must be logged on the
        error level.
        """
        class TestException(Exception):
            pass

        logger = mock.Mock()

        ctx = self.emit_exceptions(
            logger=logger,
            always_log=True,
            propagate=False
        )

        with ctx:
            raise TestException

        self.assertTrue(logger.error.called)

    def test_exception_during_emit(self):
        """
        If an exception occurs while emitting the error event, it must be
        logged and the original exception raised (if propagate is True).
        """
        class TestException(Exception):
            pass

        emitter = mock.Mock()
        logger = mock.Mock()

        emitter.emit.side_effect = RuntimeError()

        ctx = self.emit_exceptions(
            emitter,
            logger,
            propagate=True
        )

        with self.assertRaises(TestException):
            with ctx:
                raise TestException

        self.assertTrue(logger.exception.called)

    def test_skip_exception_during_emit(self):
        """
        If an exception occurs while emitting the error event, AND it matches
        the supplied ``skip_types`` it must NOT be logged.
        """
        class TestException(Exception):
            pass

        emitter = mock.Mock()
        logger = mock.Mock()

        emitter.emit.side_effect = RuntimeError()

        ctx = self.emit_exceptions(
            emitter,
            logger,
            skip_types=(RuntimeError,)
        )

        with self.assertRaises(TestException):
            with ctx:
                raise TestException

        self.assertFalse(logger.exception.called)



"""
Tests for `biloba.service`.
"""

import unittest
import mock

import gevent

from biloba import service


class SimpleService(service.Service):
    """
    Simple test service
    """
    def __init__(self):
        super(SimpleService, self).__init__()

    def do_start(self):
        """
        Start a simple service.
        """
        self.spawn(self.sleep)

    def sleep(self):
        """
        Sleep forever.
        """
        while True:
            gevent.sleep(1)


def make_service(logger=None):
    my_service = service.Service()

    if logger:
        my_service.logger = logger

    return my_service


class ServiceTestCase(unittest.TestCase):
    """
    Tests for `service.Service`
    """

    def test_create(self):
        """
        Ensure basic attributes when initialising the service
        """
        from gevent import pool

        my_service = make_service()

        self.assertFalse(my_service.started)
        self.assertEqual(my_service.services, [])
        self.assertIsInstance(my_service.pool, pool.Group)

    def test_get_logger(self):
        """
        A service should provide a logger as an attribute.
        """
        import logbook

        my_service = make_service()

        logger = my_service.logger

        self.assertIsInstance(logger, logbook.Logger)
        self.assertEqual(logger.name, 'Service')

    def test_get_logger_custom_name(self):
        """
        A service logger should be able to be named.
        """
        import logbook

        class MyService(service.Service):
            logger_name = 'my_log_name'

        my_service = MyService()

        logger = my_service.logger

        self.assertIsInstance(logger, logbook.Logger)
        self.assertEqual(logger.name, 'my_log_name')

    def test_start(self):
        """
        Ensure that starting a service works correctly.
        """
        my_service = SimpleService()

        self.event_fired = False

        def on_start():
            self.event_fired = True

        my_service.on('start', on_start)

        self.assertFalse(my_service.started)

        my_service.start()

        self.assertTrue(my_service.started)
        self.assertTrue(self.event_fired)

        # now test calling `start` again. It should be idempotent
        self.event_fired = False

        my_service.start()

        self.assertFalse(self.event_fired)

    @mock.patch.object(service.Service, 'do_start')
    def test_do_start(self, mock_do_start):
        """
        Ensure that do_start is called during `start`.
        """
        my_service = make_service()

        # Run something on service so it doesn't stop right away.
        my_service.spawn(gevent.sleep, 1)

        my_service.start()

        mock_do_start.assert_called_once_with()

        # ensure that `do_start` is called only once.
        my_service.start()

        mock_do_start.assert_called_once_with()

    def test_start_services(self):
        """
        Any greenlets in `service.services` must be `start`ed.
        """
        my_service = make_service(logger=mock.Mock())
        mock_greenlet = mock.Mock()

        my_service.add_service(mock_greenlet)

        my_service.join()

        mock_greenlet.join.assert_called_once_with()

    @mock.patch.object(service.Service, 'stop')
    def test_start_emit_error(self, mock_stop):
        """
        When a service is starting, the start event is emitted. If an exception
        occurs when emitting that event, the service must be torn down.
        """
        my_service = make_service(logger=mock.Mock())

        @my_service.on('start')
        def on_start():
            raise RuntimeError()

        with self.assertRaises(RuntimeError):
            my_service.start()

        self.assertTrue(mock_stop.called)

    @mock.patch.object(service.Service, 'teardown_service')
    def test_do_start_error(self, mock_teardown):
        """
        If an exception is raised when calling ``do_start`` while the service
        is starting, it must be torn down.
        """
        class MyService(service.Service):
            def do_start(self):
                raise RuntimeError

        my_service = MyService()
        my_service.logger = mock.Mock()

        with self.assertRaises(RuntimeError):
            my_service.start()

        mock_teardown.assert_called_once_with()

    def test_stop_not_started(self):
        """
        Can only `stop` a service when it has been started.
        """
        my_service = make_service()
        self.event_fired = False

        def on_stop():
            self.event_fired = True

        my_service.on('stop', on_stop)

        my_service.stop()

        self.assertFalse(self.event_fired)

    def test_stop(self):
        """
        Ensure that `stop`ping a service works correctly.
        """
        my_service = SimpleService()

        my_service.start()

        self.event_fired = False

        def on_stop():
            self.event_fired = True

        my_service.on('stop', on_stop)

        self.assertTrue(my_service.started)

        my_service.stop()

        self.assertFalse(my_service.started)
        self.assertTrue(self.event_fired)

        # now test calling `stop` again. It should be idempotent
        self.event_fired = False

        my_service.stop()

        self.assertFalse(self.event_fired)

    @mock.patch.object(service.Service, 'do_stop')
    def test_do_stop(self, mock_do_stop):
        """
        Ensure that `do_stop` is called during `stop`.
        """
        my_service = SimpleService()

        my_service.start()

        my_service.stop()

        mock_do_stop.assert_called_once_with()

        # ensure that `do_stop` is called only once.
        my_service.stop()

        mock_do_stop.assert_called_once_with()

    def test_stop_services(self):
        """
        Any greenlets in `service.services` must be `stop`ped.
        """
        my_service = SimpleService()
        my_service.start()

        mock_greenlet = mock.Mock()

        my_service.services = [mock_greenlet]

        my_service.stop()

        self.assertEqual(my_service.services, [])

        mock_greenlet.stop.assert_called_once_with()

    @mock.patch.object(service.Service, 'stop')
    @mock.patch.object(service.Service, 'start')
    def test_join(self, mock_start, mock_stop):
        """
        Ensure that `join` works as one would expect.
        """
        my_service = make_service()
        thread = my_service._run_thread = mock.Mock()

        my_service.join()

        mock_start.assert_called_once_with()
        thread.get.assert_called_once_with()
        mock_stop.assert_called_once_with()

    def test_spawn_error(self):
        """
        Spawning a greenlet that raises an exception must emit the error event
        via the service.
        """
        my_service = make_service()

        self.executed = False

        def trap_error(exc_type, exc_value, exc_traceback):
            self.assertIsInstance(exc_value, RuntimeError)
            self.executed = True

        def raise_error():
            raise RuntimeError

        my_service.on('error', trap_error)

        my_service.spawn(raise_error)

        gevent.sleep(0.0)

        self.assertTrue(self.executed)

    def test_spawn_exit(self):
        """
        Raising `gevent.GreenletExit` is not an error as such and the thread
        should exit correctly without emitting an `error` event.
        """
        my_service = make_service()

        self.executed = False

        def trap_error(exc_type, exc_value, exc_traceback):
            raise AssertionError('Whoops the error was trapped')

        def raise_error():
            self.executed = True
            raise gevent.GreenletExit

        my_service.on('error', trap_error)

        my_service.spawn(raise_error)

        gevent.sleep(0.0)

        self.assertTrue(self.executed)

    @mock.patch.object(service.Service, 'spawn')
    def test_add_service(self, mock_spawn):
        """
        Adding a child service when the parent service is stopped.
        """
        my_service = make_service()
        mock_service = mock.Mock()

        my_service.add_service(mock_service)

        self.assertFalse(mock_spawn.called)

        self.assertEqual(my_service.services, [mock_service])

    def test_service_error(self):
        """
        If a child service emits an error, the parent service must receive it.
        """
        child = make_service()
        parent = make_service()

        parent.add_service(child)

        self.executed = False

        @parent.on('error')
        def my_error(exc_type, exc_value, exc_tb):
            self.assertIs(exc_type, RuntimeError)
            self.assertIsInstance(exc_value, RuntimeError)
            self.assertIsNone(exc_tb)

            self.executed = True

        child.emit('error', RuntimeError())

        self.assertTrue(self.executed)

    @mock.patch.object(service.Service, 'watch_service')
    def test_add_service_start(self, mock_watch):
        """
        Adding a child service when the parent service is started.
        """
        mock_watch.__name__ = 'watch_service'

        my_service = SimpleService()
        mock_service = mock.Mock()

        my_service.start()

        my_service.add_service(mock_service)
        gevent.sleep(0.0)

        mock_watch.assert_called_once_with(mock_service)
        self.assertEqual(my_service.services, [mock_service])

    def test_teardown(self):
        """
        Teardown must be called before the state of the service is torn down.
        """
        class MyService(service.Service):
            def do_teardown(self):
                self.test.assertTrue(self.started)
                self.executed = True

        my_service = MyService()
        my_service.test = self

        my_service.start()
        my_service.stop()

        self.assertTrue(my_service.executed)

    def test_teardown_not_started(self):
        """
        Teardown must not be called if the service has not been started.
        """
        class MyService(service.Service):
            executed = False

            def do_teardown(self):
                self.executed = True

        my_service = MyService()

        my_service.stop()

        self.assertFalse(my_service.executed)

    def test_teardown_exception(self):
        """
        If there is an exception in `do_teardown`, the service must still be
        torn down.
        """
        self.executed = False

        class MyService(service.Service):
            def do_teardown(self):
                raise RuntimeError

        def check_error(exc_type, exc_value, exc_tb):
            self.assertIsInstance(exc_value, RuntimeError)
            self.executed = True

        my_service = MyService()
        mock_service = mock.Mock()

        my_service.add_service(mock_service)

        my_service.start()

        my_service.on('error', check_error)

        # stopping the service will not raise the exception but will dump it
        # to the logger
        my_service.stop()

        mock_service.stop.assert_called_once_with()
        self.assertTrue(self.executed)

    def test_no_greenlets_or_child_services(self):
        """
        If there are no greenlets spawned on the service or child services
        added, the service should start then immediately stop.
        """
        self.started = False
        self.stopped = False

        def start():
            self.started = True

        def stop():
            self.stopped = True

        class MyService(service.Service):
            def do_start(self):
                start()

            def do_stop(self):
                stop()

        my_service = MyService()

        my_service.start()

        gevent.sleep(0)

        self.assertTrue(self.started)
        self.assertTrue(self.stopped)
        self.assertFalse(my_service.started)

    def test_start_stop_simple_service(self):
        """
        If the service spawns a greenlet that runs forever, it should run until
        it is stopped.
        """
        self.executed = False

        def run(time):
            self.executed = True
            gevent.sleep(time)

        class MyService(service.Service):
            def do_start(self):
                self.spawn(run, 10)

        my_service = MyService()

        my_service.start()

        self.assertTrue(self.executed)
        self.assertTrue(my_service.started)

        my_service.stop()

        self.assertFalse(my_service.started)

    def test_start_stop_simple_service2(self):
        """
        If the service spawns a single greenlet, it should run until that
        greenlet has finished.
        """
        self.executed = False

        def run(time):
            self.executed = True
            gevent.sleep(time)

        class MyService(service.Service):
            def do_start(self):
                self.spawn(run, 0)

        my_service = MyService()

        my_service.start()

        self.assertTrue(self.executed)
        self.assertTrue(my_service.started)

        my_service.join()

        self.assertFalse(my_service.started)

    def test_start_stop_with_child_service(self):
        """
        If the service has a child service that runs forever, it should run
        until it is stopped.
        """
        self.executed = False

        def run(time):
            self.executed = True
            gevent.sleep(time)

        class ChildService(service.Service):
            def do_start(self):
                self.spawn(run, 10)

        my_service = make_service()
        child_service = ChildService()

        my_service.add_service(child_service)

        my_service.start()

        self.assertTrue(self.executed)
        self.assertTrue(my_service.started)
        self.assertTrue(child_service.started)

        my_service.stop()

        self.assertFalse(my_service.started)
        self.assertFalse(child_service.started)

    def test_start_stop_with_child_service2(self):
        """
        If the service has a single child service, it should run until that
        service has stopped.
        """
        self.executed = False

        def run(time):
            self.executed = True
            gevent.sleep(time)

        class ChildService(service.Service):
            def do_start(self):
                self.spawn(run, 0)

        my_service = make_service()
        child_service = ChildService()

        my_service.add_service(child_service)

        my_service.start()

        self.assertTrue(self.executed)
        self.assertTrue(my_service.started)
        self.assertTrue(child_service.started)

        my_service.join()

        self.assertFalse(my_service.started)
        self.assertFalse(child_service.started)

    def test_stopping_with_a_child_service(self):
        """
        Calling stop on a service that has a child service should not cause
        a `LoopExit` exception.
        """
        self.executed = [False, False]

        def run(index, time):
            self.executed[index] = True
            gevent.sleep(time)

        class MyService(service.Service):
            def do_start(self):
                self.spawn(run, 0, 0)

        class ChildService(service.Service):
            def do_start(self):
                self.spawn(run, 1, 10)

        my_service = MyService()
        child_service = ChildService()

        my_service.add_service(child_service)

        my_service.start()

        self.assertTrue(all(self.executed))
        self.assertTrue(my_service.started)
        self.assertTrue(child_service.started)

        my_service.stop()

    def test_stopping_with_a_child_service2(self):
        """
        A service that has a child service should be able to stop itself
        without causing a `LoopExit` exception.
        """
        self.executed = [False, False]

        def run(index, time):
            self.executed[index] = True
            gevent.sleep(time)

        class MyService(service.Service):
            def do_start(self):
                def run_and_stop():
                    run(0, 0)
                    gevent.spawn(self.stop)

                self.spawn(run_and_stop)

        class ChildService(service.Service):
            def do_start(self):
                self.spawn(run, 1, 10)

        my_service = MyService()
        child_service = ChildService()

        my_service.add_service(child_service)

        my_service.start()

        self.assertTrue(all(self.executed))
        self.assertTrue(my_service.started)
        self.assertTrue(child_service.started)

        my_service.join()


class ConfigurableServiceTestCase(unittest.TestCase):
    """
    Tests for `service.ConfigurableService`.
    """

    def test_create(self):
        """
        Create a configurable service.
        """
        from biloba import config

        my_service = service.ConfigurableService(None)

        self.assertIsInstance(my_service.config, config.Config)

    def test_default_config(self):
        """
        Test default config
        """
        class MyService(service.ConfigurableService):
            def get_config_defaults(self):
                return {'foo': 'bar'}

        my_service = MyService(None)

        self.assertEqual(my_service.config['foo'], 'bar')

    def test_override_default_config(self):
        class MyService(service.ConfigurableService):
            def get_config_defaults(self):
                return {'foo': 'bar'}

        my_service = MyService({'foo': 'baz'})

        self.assertEqual(my_service.config['foo'], 'baz')

    def test_biloba_config(self):
        """
        Supplying a :ref:`biloba.config.Config` must set that as the config
        instance for the service.
        """
        from biloba import config

        cfg = config.Config()

        svc = service.ConfigurableService(cfg)

        self.assertIs(svc.config, cfg)
